library geiger_localstorage;

import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:intl/locale.dart';

class ParserException extends StorageException {
  ParserException(String message,
      [StackTrace? stackTrace, Exception? cause, StackTrace? causeStackTrace])
      : super(message, stackTrace, cause, causeStackTrace);
}

class NodeParser {
  int _line = 0;

  int _character = 0;

  String _buffer;

  Future<NodeParser> fromByteStream(ByteStream stream,
      {int line = 0, int character = 0}) async {
    return NodeParser(await stream.convertToString(), line, character);
  }

  NodeParser(this._buffer, [this._line = 1, this._character = 0]);

  void updateCounter(String extract) {
    for (String i in extract.split('')) {
      if (i != '\n') {
        _character++;
      } else {
        _character = 0;
        _line++;
      }
    }
  }

  String getString([int length = 1]) {
    String ret = peekString(length);
    _buffer = _buffer.substring(length);
    updateCounter(ret);
    return ret;
  }

  String peekString([int length = 1]) {
    if (_buffer.length < length) {
      return _buffer;
    } else {
      return _buffer.substring(0, length);
    }
  }

  String getQuotedString() {
    String ret = '';
    String quoteChar = peekString(1);

    if (quoteChar != "'" && quoteChar != '"') {
      throw ParserException('expected quotes at line $line position $pos');
    }
    skipChar(quoteChar);
    while (peekString() != quoteChar) {
      if (peekString() == '\\') {
        skipChar('\\');
        ret += getString(1);
      } else {
        ret += getString(1);
      }
    }
    skipChar(quoteChar);
    return ret;
  }

  void skipChar(String char) {
    if (char == '') {
      getString(1);
    } else {
      for (String i in char.split('')) {
        if (getString(1) != i) {
          throw ParserException('expected "$char" at line $line position $pos');
        }
      }
    }
  }

  void skipWhitespaces() {
    while (_isWhitespace(peekString())) {
      skipChar('');
    }
  }

  void skipNewlines() {
    while (_isNewline(peekString())) {
      skipChar('');
    }
    skipWhitespaces();
  }

  void skipNewlineAndWhitespaces() {
    while (_isNewline(peekString()) || _isWhitespace(peekString())) {
      skipChar('');
    }
    skipWhitespaces();
  }

  static const String _atomChar =
      'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_.-';

  static const String _pathChar = '$_atomChar:';

  static bool _isWhitespace(String char) => char == '\t' || char == ' ';

  static bool _isNewline(String char) => char == '\r' || char == '\n';

  static bool _isPathChar(String char) => _pathChar.contains(char);

  static bool _isAtomChar(String char) => _atomChar.contains(char);

  int get line => _line;
  int get pos => _character;

  NodeValue getNodeValue() {
    String name = getAtom();
    Map<Locale, String> localizations = <Locale, String>{};
    skipNewlineAndWhitespaces();
    skipChar('=');
    skipNewlineAndWhitespaces();
    skipChar('{');
    skipNewlineAndWhitespaces();
    // get localizations
    while (peekString() != '}') {
      Locale loc = NodeValueImpl.defaultLocale;
      try {
        loc = Locale.parse(getAtom());
      } on FormatException {
        throw ParserException('Illegal locale at line $line pos $pos');
      }
      skipNewlineAndWhitespaces();
      skipChar('=>');
      skipNewlineAndWhitespaces();
      String val = getQuotedString();
      skipNewlineAndWhitespaces();
      if (peekString() != '}') {
        skipChar(',');
        skipNewlineAndWhitespaces();
        if (peekString() == '}') {
          throw ParserException('atom expected at line $line pos $pos');
        }
      }
      localizations[loc] = val;
    }
    skipNewlineAndWhitespaces();
    skipChar('}');
    NodeValue ret =
        NodeValueImpl(name, localizations[NodeValueImpl.defaultLocale]!);
    for (MapEntry<Locale, String> me in localizations.entries) {
      ret.setValue(me.value, me.key);
    }
    return ret;
  }

  Map<Field, String> getOrdinals() {
    Map<Field, String> ret = <Field, String>{};
    skipWhitespaces();
    if (peekString() == '[') {
      // get ordinals
      skipChar('[');
      while (_buffer != '' && peekString() != ']') {
        // get next ordinal
        skipWhitespaces();
        String ordinal = getAtom();
        Field? ordinalf = Field.fromString(ordinal);
        if (ordinalf == null) {
          throw ParserException(
              'expected valid ordinal name at line $line pos $pos (got "$ordinal")');
        }
        skipWhitespaces();
        skipChar('=');
        String value = getAtom();
        ret[ordinalf] = value;
        skipNewlineAndWhitespaces();
        if (peekString() != ']') {
          skipChar(',');
          skipNewlineAndWhitespaces();
          if (peekString() == ']') {
            throw ParserException('ordinal expected at line $line pos $pos');
          }
        }
      }
      skipChar(']');
    }
    return ret;
  }

  Future<Node> getNode() async {
    skipNewlineAndWhitespaces();
    String path = getPath();
    skipWhitespaces();
    Map<Field, String> ordinals = getOrdinals();
    String? timestamp = ordinals[Field.lastModified];
    String? owner = ordinals[Field.owner];
    if (owner == null) {
      throw ParserException(
          'owner ordinal may not be left out at line $line pos $pos');
    }
    NodeImpl ret = NodeImpl(path, owner);
    for (Field f in ordinals.keys) {
      await ret.set(f, ordinals[f]);
    }
    skipNewlineAndWhitespaces();
    // get values
    skipChar('{');
    skipNewlineAndWhitespaces();
    while (peekString() != '}') {
      skipNewlineAndWhitespaces();
      // get value
      await ret.addValue(getNodeValue());
      skipNewlineAndWhitespaces();
      if (peekString() != '}') {
        skipChar(',');
        skipNewlineAndWhitespaces();
        if (peekString() == '}') {
          throw ParserException('NodeValue expected at line $line pos $pos');
        }
      }
    }
    skipChar('}');
    skipNewlineAndWhitespaces();
    // get children
    if (peekString(8) == 'children') {
      skipChar('children');
      skipNewlineAndWhitespaces();
      skipChar('{');
      while (peekString() != '}') {
        String name =
            '${ret.path}${GenericController.pathDelimiter}${getPath()}';
        try {
          ret.addChild(
              NodeImpl.createSkeleton(name, ret.owner, ret.controller));
        } on StorageException catch (e, st) {
          throw ParserException(
              'Error while creating skeletonized children (path: ${ret.path}; name: ${getPath()}) at $line pos $pos',
              null,
              e,
              st);
        }
        skipNewlineAndWhitespaces();
        if (peekString(1) == ',') {
          skipChar(',');
          if (peekString() == '}') {
            throw ParserException(
                'expected children name at line $line pos $pos');
          }
          skipNewlineAndWhitespaces();
        }
      }
      skipChar('}');
      skipNewlineAndWhitespaces();
    }

    // make sure that timestamp remains unaltered
    await ret.set(Field.lastModified, timestamp);
    return ret;
  }

  String getPath() {
    String ret = '';
    while (_isPathChar(peekString())) {
      ret += getString();
    }
    return ret;
  }

  String getAtom() {
    String ret = '';
    while (_isAtomChar(peekString())) {
      ret += getString();
    }
    return ret;
  }

  Future<List<Node>> getNodes() async {
    List<Node> ret = <Node>[];
    skipNewlineAndWhitespaces();
    while (_buffer != '') {
      ret.add(await getNode());
      skipNewlineAndWhitespaces();
    }
    return ret;
  }
}
