// ignore_for_file: prefer_const_constructors, deprecated_member_use_from_same_package

library geiger_localstorage;

import 'dart:async';
import 'dart:collection';
import 'dart:io';

import 'package:flutter/services.dart' show rootBundle;
import 'package:flutter/widgets.dart' as widget;
import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:logging/logging.dart';
import 'package:uuid/uuid.dart';

/// This Class Acts as an intermediate class to relay storageRequests to the StorageMapper.
///
/// This class contains the interaction API every actor (sensor, shields, UI,...)
/// calls methods from this controller to manipulate data.
class GenericController implements StorageController, ChangeRegistrar {
  static final Logger _logger = Logger('GeigerStorage');

  static StorageController? defaultController;

  @Deprecated(
      'DO NOT USE IN PRODUCTION CODE. Feature will be removed in the upcomming 0.7 version.')
  static bool preloadData = true;

  /// the path delimiter.
  static const String pathDelimiter = ':';

  /// the default owner of all newly created nodes.
  final String _owner;

  /// a map of all registered listeners for changes.
  final Map<SearchCriteria, Set<StorageListener>> listeners = HashMap();

  /// the storage mapper to be used.
  final StorageMapper mapper;

  int _inited = 0;

  /// <p>Construct the controller according to the database that is used.</p>
  ///
  /// @param owner  the owner to be used for all requests
  /// @param mapper the database mapper to use
  ///
  /// @throws StorageException In case of problems regarding the storage
  GenericController(this._owner, this.mapper) {
    mapper.setController(this);
    defaultController = this;
  }

  /// <p>Returns the latest created controller within the JVM.</p>
  ///
  /// @return a valid storage controller
  static StorageController? getDefault() {
    return defaultController;
  }

  static Future<String> getResourceOrFile(String? name) async {
    if (!preloadData) {
      return '';
    }
    try {
      if (name == null) {
        throw Exception('no name passed... falling back to internal assets');
      }
      File f = File(name);
      return await f.readAsString();
    } catch (e) {
      _logger.log(Level.INFO,
          'no local file found named $name... falling back to resource', e);
      try {
        String? ret;
        // searching resources
        widget.WidgetsFlutterBinding.ensureInitialized();
        for (String s in [
          'assets/store/store.data',
          'lib/assets/store/store.data',
          'packages/geiger_localstorage/assets/store/store.data',
          'packages/geiger_localstorage/lib/assets/store/store.data'
        ]) {
          try {
            ret ??= await rootBundle.loadString(s);
          } catch (e) {
            _logger.log(Level.INFO, 'no resource file found for $s');
          }
          try {
            File f = File(s);
            ret ??= await f.readAsString();
          } catch (e) {
            _logger.log(Level.INFO, 'No file found for $s');
          }
        }
        if (ret == null) {
          _logger.log(Level.WARNING,
              'no resource file found... out of options for preloading file');
          return '';
        } else {
          _logger.log(Level.INFO, 'resource file found');
          return ret;
        }
      } catch (e) {
        _logger.log(
            Level.WARNING,
            'no ressource file found... out of options for preloading file (2)',
            e);
        return '';
      }
    }
  }

  Future<void> _initMapper() async {
    if (_inited == 0) {
      _inited = 1;
      final baseNodes = <String>[
        pathDelimiter,
        '${pathDelimiter}Devices',
        '${pathDelimiter}Users',
        '${pathDelimiter}Enterprise',
        '${pathDelimiter}Keys',
        '${pathDelimiter}Global',
        '${pathDelimiter}Local'
      ];
      for (String nodeName in baseNodes) {
        try {
          // for correct path generation creation of a Node is needed
          await _get(nodeName);
        } on StorageException {
          // node does not exist and therefore we create it
          Node tmp = await NodeImpl.fromPath(nodeName, _owner);
          tmp.visibility = Visibility.white;
          tmp.owner = _owner;
          await _add(tmp);
        }
      }

      // check if current user exists
      NodeValue? uuid =
          await (_getValue('${pathDelimiter}Local', 'currentUser'));
      if (uuid == null) {
        uuid = NodeValueImpl('currentUser', Uuid().v4());
        await _addValue('${pathDelimiter}Local', uuid);
      }

      // check if current user node exists
      String userNodeName = '${pathDelimiter}Users$pathDelimiter' + uuid.value;
      try {
        await _get(userNodeName);
      } on StorageException {
        Node n = await NodeImpl.fromPath(userNodeName, _owner);
        await _add(n);
      }

      // check if current device exists
      uuid = await _getValue('${pathDelimiter}Local', 'currentDevice');
      if (uuid == null) {
        // create new default device
        uuid = NodeValueImpl('currentDevice', Uuid().v4());
        await _addOrUpdateValue('${pathDelimiter}Local', uuid);
      }

      // check if current device node exists
      var deviceNodeName = '${pathDelimiter}Devices$pathDelimiter' + uuid.value;
      try {
        await _get(deviceNodeName);
      } on StorageException {
        await _add(await NodeImpl.fromPath(deviceNodeName, _owner));
      }

      // add all initial nodes
      widget.WidgetsFlutterBinding.ensureInitialized();

      String initer = await getResourceOrFile('lib/assets/store/store.data');
      NodeParser np = NodeParser(initer);
      List<Node> nl = await np.getNodes();
      for (Node n in nl) {
        try {
          // check if node exists
          await _getNodeOrTombstone(n.path);
          _logger.log(
              Level.FINE, "skipping node ${n.path} as it already exists.");
        } on StorageException {
          try {
            await _add(n);
          } on StorageException catch (e, st) {
            _logger.log(Level.SEVERE,
                "unable to add node ${n.path} while initializing.", e, st);
          }
        }
      }

      _inited = 2;
      await ExtendedTimestamp.initializeTimestamp(this);
    } else if (_inited == 1) {
      await Future.doWhile(() => _inited == 2);
    }
  }

  @override
  Future<bool> addOrUpdate(Node? node) async {
    await _initMapper();
    return await _addOrUpdate(node);
  }

  Future<bool> _addOrUpdate(Node? node) async {
    if (node == null) {
      return false;
    } else if (node.skeleton) {
      return false;
    } else if (node.tombstone) {
      await delete(node.path);
      return false;
    } else {
      bool ret = false;
      try {
        await _add(node);
        ret = true;
      } on StorageException catch (_) {
        await _update(node);
      }
      List<Node> children = List.from((await node.getChildren()).values);
      for (Node n2 in children) {
        ret |= await _addOrUpdate(n2);
      }
      return ret;
    }
  }

  @override
  Future<Node> get(String path) async {
    await _initMapper();
    return await _get(path);
  }

  Future<Node> _get(String path) async {
    Node n = await _getNodeOrTombstone(path);
    if (n.tombstone) {
      throw StorageException(
          'Node "$path" does not exist (but exited in the past; It is a tombstone now)');
    }
    List<String> l = <String>[];
    for (Node cn in (await n.getChildren()).values) {
      if (cn.tombstone) {
        l.add(cn.name);
      }
    }
    for (String name in l) {
      await n.removeChild(name);
    }
    return n;
  }

  @override
  Future<Node> getNodeOrTombstone(String path) async {
    await _initMapper();
    return await _getNodeOrTombstone(path);
  }

  Future<Node> _getNodeOrTombstone(String path) async {
    return mapper.get(path);
  }

  @override
  Future<void> add(Node node) async {
    await _initMapper();
    return await _add(node);
  }

  Future<void> _add(Node node) async {
    // make sure that there is an owner set
    if ('' == node.owner) {
      node.owner = _owner;
    }

    // check for tombstone
    try {
      Node tombstone = await mapper.get(node.path);
      if (tombstone.tombstone) {
        await mapper.delete(node.path);
      }
    } on StorageException {
      // no tombstone to delete
    }

    // add object
    await mapper.add(node);

    // add child nodes
    for (Node n in (await node.getChildren()).values) {
      if (!n.skeleton && !n.tombstone) {
        await mapper.add(n);
      }
    }

    unawaited(_checkListeners(EventType.create, null, node, null, null));
  }

  @override
  Future<void> update(Node node) async {
    await _initMapper();
    return await _update(node);
  }

  Future<void> _update(Node node) async {
    // make sure that there is an owner set
    if ('' == node.owner) {
      node.owner = _owner;
    }

    // get old node for update events
    Node oldNode = await mapper.get(node.path);

    // write node
    await mapper.update(node);

    unawaited(_checkListeners(EventType.update, oldNode, node, null, null));

    // any child that is not a skeleton will be handled as new or changed
    for (Node child in (await node.getChildren()).values) {
      if (!child.skeleton) {
        try {
          await _add(child);
        } on StorageException {
          // node already exists, therefore it was changed
          await _update(child);
        }
      }
    }
  }

  @override
  Future<Node> delete(String path) async {
    await _initMapper();
    return await _delete(path);
  }

  Future<Node> _delete(String path) async {
    Node ret = await mapper.delete(path);
    unawaited(_checkListeners(EventType.delete, ret, null, null, null));
    return ret;
  }

  @override
  Future<void> rename(String oldPath, String newPathOrName) async {
    await _initMapper();
    return await _rename(oldPath, newPathOrName);
  }

  Future<void> _rename(String oldPath, String newPathOrName) async {
    Node oldNode = await mapper.get(oldPath);
    String newPath = newPathOrName;
    if (!newPathOrName.startsWith(pathDelimiter)) {
      // create path from name
      newPath = (oldNode.parentPath!.endsWith(':')
              ? oldNode.parentPath!.substring(0, oldNode.parentPath!.length - 1)
              : oldNode.parentPath!) +
          pathDelimiter +
          newPathOrName;
    }
    await mapper.rename(oldPath, newPath);
    Node newNode = await mapper.get(newPath);
    unawaited(_checkListeners(EventType.rename, oldNode, newNode, null, null));
  }

  @override
  Future<NodeValue?> getValue(String path, String key) async {
    await _initMapper();
    return await _getValue(path, key);
  }

  Future<NodeValue?> _getValue(String path, String key) async {
    return await mapper.getValue(path, key);
  }

  @override
  Future<void> addValue(String nodeName, NodeValue newValue) async {
    await _initMapper();
    return await _addValue(nodeName, newValue);
  }

  Future<void> _addValue(String nodeName, NodeValue newValue) async {
    Node oldNode = await mapper.get(nodeName);
    NodeValue? oldValue = await oldNode.getValue(newValue.key);
    if (oldValue != null) {
      throw StorageException('value "' + newValue.key + '" is already set');
    }
    Node newNode = await oldNode.deepClone();
    await newNode.addValue(newValue);
    await mapper.update(newNode);
    unawaited(
        _checkListeners(EventType.update, oldNode, newNode, null, newValue));
  }

  @override
  Future<bool> addOrUpdateValue(String nodeName, NodeValue newValue) async {
    await _initMapper();
    return await _addOrUpdateValue(nodeName, newValue);
  }

  Future<bool> _addOrUpdateValue(String nodeName, NodeValue newValue) async {
    Node oldNode = await mapper.get(nodeName);
    Node newNode = await oldNode.deepClone();
    NodeValue? nv = await newNode.getValue(newValue.key);
    bool ret = false;
    if (nv == null) {
      await newNode.addValue(newValue);
      ret = true;
    } else {
      await newNode.updateValue(newValue);
    }
    await mapper.update(newNode);
    unawaited(
        _checkListeners(EventType.update, oldNode, newNode, null, newValue));
    return ret;
  }

  @override
  Future<void> updateValue(String nodeName, NodeValue newValue) async {
    await _initMapper();
    return await _updateValue(nodeName, newValue);
  }

  Future<void> _updateValue(String nodeName, NodeValue newValue) async {
    Node oldNode = await mapper.get(nodeName);
    NodeValue? oldValue = await oldNode.getValue(newValue.key);
    if (oldValue == null) {
      throw StorageException('value "' + newValue.key + '" does not yet exist');
    }
    Node newNode = await oldNode.deepClone();
    await newNode.removeValue(newValue.key);
    await newNode.addValue(newValue);
    await mapper.update(newNode);
    unawaited(_checkListeners(
        EventType.update, oldNode, newNode, oldValue, newValue));
  }

  @override
  Future<NodeValue> deleteValue(String nodeName, String key) async {
    await _initMapper();
    return await _deleteValue(nodeName, key);
  }

  Future<NodeValue> _deleteValue(String nodeName, String key) async {
    Node oldNode = await mapper.get(nodeName);
    NodeValue? oldValue = await oldNode.getValue(key);
    if (oldValue == null) {
      throw StorageException('value "' + key + '" does not yet exist');
    }
    Node newNode = await oldNode.deepClone();
    await newNode.removeValue(key);
    await mapper.update(newNode);
    unawaited(
        _checkListeners(EventType.update, oldNode, newNode, oldValue, null));
    return oldValue;
  }

  @override
  Future<List<Node>> search(SearchCriteria criteria) async {
    return await mapper.search(criteria);
  }

  @override
  Future<void> close() async {
    await mapper.close();
  }

  @override
  Future<void> flush() async {
    await _initMapper();
    await mapper.flush();
  }

  Future<void> _checkListeners(final EventType event, final Node? oldNode,
      final Node? newNode, NodeValue? oldValue, NodeValue? newValue) async {
    if (oldNode == null || oldNode != newNode) {
      List<MapEntry<SearchCriteria, Set<StorageListener>>> me =
          listeners.entries.toList();
      for (MapEntry<SearchCriteria, Set<StorageListener>> e in me) {
        try {
          if ((oldNode != null &&
                  newNode != null &&
                  (await e.key.evaluate(oldNode) ||
                      await e.key.evaluate(newNode))) ||
              (oldNode != null && await e.key.evaluate(oldNode)) ||
              (newNode != null && await e.key.evaluate(newNode))) {
            try {
              for (StorageListener l in e.value) {
                l.gotStorageChange(event, oldNode, newNode);
              }
            } catch (e) {
              // TODO(unassigned): raise sensible exception
              // FIXME do something sensible (should not happen anyway)
              rethrow;
            }
          }
        } on StorageException {
          // TODO(unassigned): raise sensible exception
          // FIXME do something sensible (should not happen anyway)
        }
      }
      // }
    }
  }

  @override
  Future<void> registerChangeListener(
      StorageListener listener, SearchCriteria criteria) async {
    // synchronized(listeners) {
    if (listeners[criteria] == null) {
      listeners[criteria] = <StorageListener>{};
    }
    listeners[criteria]!.add(listener);
    // }
  }

  @override
  Future<List<SearchCriteria>> deregisterChangeListener(
      StorageListener listener) async {
    List<SearchCriteria> remove = <SearchCriteria>[];
    for (MapEntry<SearchCriteria, Set<StorageListener>> e
        in listeners.entries) {
      if (e.value.contains(listener)) {
        remove.add(e.key);
      }
    }
    if (remove.isEmpty) {
      throw StorageException('Listener not registered');
    }
    for (var c in remove) {
      listeners[c]!.remove(listener);
    }
    return remove;
  }

  @override
  Future<void> zap() async {
    ExtendedTimestamp.reset();
    await mapper.zap();
    _inited = 0;
    ExtendedTimestamp.reset();
    // init database
    await _initMapper();
    // make sure that Extended Timestamp is inited again
    await ExtendedTimestamp.initializeTimestamp(this);
  }

  @override
  Future<String> dump([String rootNode = ':', String prefix = '']) async {
    NodeImpl n = (await getNodeOrTombstone(rootNode)) as NodeImpl;
    String ret = n.toString(prefix: prefix);
    for (Node nc in (await n.getChildren()).values) {
      ret = '$ret\n${await dump(nc.path, '$prefix   ')}';
    }
    return ret;
  }
}
