library geiger_localstorage;

import 'dart:collection';

import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:intl/locale.dart';
import 'package:logging/logging.dart';
import 'package:sqlite3/sqlite3.dart';

/// This class maps the database interface functions to a SQLite database
class SqliteMapper extends AbstractSqlMapper {
  static final Logger _logger = Logger('GeigerStorage');
  static const int maxFieldSize = 20 * 1024 * 1024;
  static final Map<String, String> _dbBuilder =
      LinkedHashMap<String, String>.fromIterable(<List<String>>[
    [
      'SELECT * FROM storage_node;\n',
      'CREATE TABLE IF NOT EXISTS storage_node (\n'
              'path TEXT CHECK( LENGTH(path) <=' +
          maxFieldSize.toString() +
          ') PRIMARY KEY,\n'
              'owner TEXT CHECK( LENGTH(owner) <= 40),\n'
              'name TEXT CHECK( LENGTH(name) <= 40) NOT NULL,\n'
              "visibility TEXT CHECK(visibility IN ('black', 'red', 'amber', 'green', 'white')) NOT NULL DEFAULT 'red',\n"
              'children TEXT CHECK( LENGTH(children) <=' +
          maxFieldSize.toString() +
          ') NULL,\n'
              "tombstone TEXT CHECK( tombstone IN ('TRUE', 'FALSE')) DEFAULT 'FALSE'\n"
              ');\n'
    ],
    [
      'SELECT * FROM node_value;\n',
      'CREATE TABLE IF NOT EXISTS node_value (\n'
              'path TEXT CHECK( LENGTH(path) <=' +
          maxFieldSize.toString() +
          ') NOT NULL,\n'
              'key TEXT CHECK( LENGTH(key) <= 40) NOT NULL,\n'
              'value TEXT CHECK( LENGTH(value) <= 10 * 1024 * 1024),\n'
              'type TEXT CHECK( LENGTH(type) <= 40),\n'
              'locale TEXT CHECK( LENGTH(locale) <= 10) NOT NULL,\n'
              'last_modified TEXT CHECK( LENGTH(last_modified) <= 20) NOT NULL,\n'
              'PRIMARY KEY (path, key),\n'
              'FOREIGN KEY (path) REFERENCES storage_node(path)\n'
              ');\n'
    ],
    [
      'SELECT * FROM translation;\n',
      'CREATE TABLE IF NOT EXISTS translation (\n'
              'path TEXT CHECK( LENGTH(path) <=' +
          maxFieldSize.toString() +
          ') NOT NULL,\n'
              'key TEXT CHECK( LENGTH(key) <= 40) NOT NULL,\n'
              "identifier TEXT CHECK(identifier IN ('VALUE', 'DESCRIPTION')) NOT NULL,\n"
              'locale TEXT CHECK( LENGTH(locale) <= 10) NOT NULL,\n'
              'translation TEXT CHECK( LENGTH(translation) <=' +
          maxFieldSize.toString() +
          ') NOT NULL,\n'
              'PRIMARY KEY (path, key, identifier, locale),\n'
              'FOREIGN KEY (path, key) REFERENCES node_value(path, key)\n'
              ');\n'
    ],
    // add last modified column
    [
      'SELECT last_modified FROM storage_node LIMIT 1;',
      'ALTER TABLE storage_node ADD COLUMN last_modified TEXT;\n'
    ],
    /*[
      'SELECT sql FROM sqlite_master WHERE type=\'table\' ANS name=\'storage_node\' AND ',
      ''
    ]*/
  ], key: (e) => e[0], value: (e) => e[1]);

  late Database conn;
  StorageController? controller;

  SqliteMapper(String dbPath) {
    String fullPath = StorageMapper.expandDbFileName(dbPath);
    try {
      conn = sqlite3.open(fullPath);
      _testDB();
    } on SqliteException catch (e, st) {
      _logger.log(Level.SEVERE, 'Error opening database at $fullPath', e, st);
      throw StorageException(
          'Error opening database at $fullPath (is path writeable?)',
          null,
          e,
          st);
    }
  }

  void _testDB() {
    try {
      // insert any test here for the most recent database
      for (String k in _dbBuilder.keys) {
        conn.execute(k);
      }
    } on SqliteException {
      try {
        _logger.log(Level.INFO, 'test of database failed... running init');
        _initialize();
      } on StorageException {
        rethrow;
      }
    }
  }

  void _initialize() {
    try {
      for (String s in _dbBuilder.keys) {
        try {
          conn.execute(s);
          _logger.log(Level.FINE, 'Test for "$s" successful');
        } on Exception {
          try {
            _logger.log(
                Level.INFO, 'Running DB upgrade for "$s" (${_dbBuilder[s]})');
            conn.execute(_dbBuilder[s]!);
          } on SqliteException catch (st, t) {
            throw StorageException(
                'Whoops... error while upgrading database (offending: ${_dbBuilder[s]})',
                null,
                st,
                t);
          }
        }
      }
    } on SqliteException catch (e, ste) {
      throw StorageException(
          'Error on init/upgrade of database... try deleting the database or restore an old state',
          null,
          e,
          ste);
    } on Error {
      throw StorageException(
          'Got exception while writing database... is the database file/dir writeable?');
    }
  }

  @override
  void setController(StorageController controller) {
    this.controller = controller;
  }

  @override
  Future<Node> get(String path) async {
    checkPath(path);
    getSanity(path);
    String sqlStatement =
        'SELECT path, owner, name, visibility, children, tombstone,last_modified '
        'FROM storage_node WHERE path = ?';
    Node res;
    int lastModified;
    try {
      ResultSet resultSet = _requestResult(sqlStatement, [path]);
      if (resultSet.isEmpty) {
        throw StorageException('Node "$path" does not exist on sqlite backend');
      } else {
        String owner = (resultSet.first['owner'] ?? '<UNSET>') as String;
        if ('true' == resultSet.first['tombstone'].toLowerCase()) {
          res = await NodeImpl.fromPath(
              resultSet.first['path'] as String, owner,
              isTombstone: true);
          res.visibility = VisibilityExtension.valueOf(
                  resultSet.first['visibility']! as String) ??
              Visibility.red;
          return res;
        }
        res = await NodeImpl.fromPath(resultSet.first['path'] as String, owner);
        res.owner = owner;
        res.visibility = VisibilityExtension.valueOf(
                resultSet.first['visibility'] as String) ??
            Visibility.red;
        String children = resultSet.first['children'] as String;
        if (!('' == children)) {
          for (String childName in children.split(',')) {
            await res.addChild(NodeImpl.createSkeleton(
                (path != ':' ? path + ':' : ':') + childName,
                owner,
                controller));
          }
        }
        lastModified = int.parse(resultSet.first['last_modified']);
      }
    } on Exception catch (exception, s) {
      if (!(exception is StorageException || exception is SqliteException)) {
        rethrow;
      }
      throw StorageException(
          'Could not retrieve node "$path"', null, exception, s);
    }
    sqlStatement = ('SELECT path, key, value, type, locale, last_modified '
        'FROM node_value WHERE path = ?');
    ResultSet resultSetValues = _requestResult(sqlStatement, [path]);
    String key;
    try {
      for (Row valueRow in resultSetValues) {
        key = valueRow['key'] as String;
        NodeValue value = NodeValueImpl(
            key,
            valueRow['value'] as String,
            valueRow['type'] as String?,
            '',
            int.parse(valueRow['last_modified'] as String));
        String sqlStatementTranslations =
            'SELECT path, key, identifier, locale, translation '
            'FROM translation WHERE (path = ? AND key = ?)';
        ResultSet resultSetTranslations =
            _requestResult(sqlStatementTranslations, [path, key]);
        try {
          for (Row translationRow in resultSetTranslations) {
            String identifier = translationRow['identifier'] as String;
            if ('VALUE' == identifier) {
              value.setValue(translationRow['translation'] as String,
                  Locale.parse(translationRow['locale'] as String));
            } else {
              if ('DESCRIPTION' == identifier) {
                value.setDescription(translationRow['translation'] as String,
                    Locale.parse(translationRow['locale'] as String));
              }
            }
          }
        } on SqliteException catch (e, st) {
          throw StorageException(
              'Could not retrieve description for node "' +
                  path +
                  '" and key "' +
                  key +
                  '"',
              null,
              e,
              st);
        }
        // GwM: Unclear what this is!
        await res.addValue(value);
      }
    } on SqliteException catch (e, st) {
      throw StorageException(
          'Could not retrieve values for node "' + path + '"', null, e, st);
    }
    // Has to be add the end because else the modification above will change it
    res.lastModified = lastModified;
    return res;
  }

  @override
  Future<void> add(Node node) async {
    // do not add skeleton nodes
    if (node.skeleton) {
      throw StorageException('Skeleton nodes cannot be added.');
    }
    checkPath(node.path);
    try {
      Node n = await get(node.path);
      if (!n.tombstone) {
        throw Exception('Node "${node.path}" already exists');
      } else {
        await delete(node.path, deleteTombStone: true);
      }
    } on Exception catch (e, st) {
      if ((e is! StorageException)) {
        throw StorageException(
            'Node "${node.path}" already exists', null, e, st);
      }
    }

    // update children
    if (node.parentPath != null) {
      String? ppath = node.parentPath != '' ? node.parentPath : ':';
      Node parent = await get(ppath!);
      await parent.addChild(node);
      await update(parent);
    }

    // add SQL table entry for node
    String sqlStatement = 'INSERT INTO '
        'storage_node(path, owner, name, visibility, children, tombstone,last_modified) '
        'VALUES (?,?,?,?,?,?,?)';
    try {
      var args = [
        node.path,
        node.owner,
        node.name,
        (node.visibility).toValueString(),
        await node.getChildNodesCsv(),
        node.tombstone.toString().toUpperCase(),
        node.lastModified
      ];
      _requestResult(sqlStatement, args);
    } on StorageException catch (e, s) {
      throw StorageException(
          'Could not add new node "${node.path}"', null, e, s);
    }

    // write node values
    Iterable<NodeValue> nvl = (await node.getValues()).values;
    for (NodeValue nv in nvl) {
      await addValue(node.path, nv);
    }
  }

  @override
  Future<void> update(Node node) async {
    if (node.skeleton) {
      throw StorageException('Skeleton nodes cannot be added, '
          'please materialize before adding');
    }
    checkPath(node.path);
    await get(node.path);
    String sqlStatement =
        'UPDATE storage_node SET owner = ?, visibility = ?, children = ?, '
        'tombstone = ?,last_modified = ? WHERE path = ?';
    try {
      _requestResult(sqlStatement, [
        node.owner,
        (node.visibility).toValueString(),
        await node.getChildNodesCsv(),
        node.tombstone.toString().toUpperCase(),
        node.lastModified,
        node.path
      ]);
      for (MapEntry<String, NodeValue> entry
          in (await node.getValues()).entries) {
        if (await getValue(node.path, entry.key) == null) {
          await addValue(node.path, entry.value);
        } else {
          await updateValue(node.path, entry.value);
        }
      }
    } on StorageException catch (e, st) {
      throw StorageException('Could not update node', null, e, st);
    }
  }

  @override
  Future<Node> delete(String path, {bool deleteTombStone = false}) async {
    var oldNode = await get(path);
    if ((await oldNode.getChildren()).isNotEmpty) {
      throw StorageException(
          'Node does have ${(await oldNode.getChildren()).length} children... cannot remove ' +
              path);
    }
    for (NodeValue nv in (await oldNode.getValues()).values) {
      await deleteValue(path, nv.key);
    }
    if (deleteTombStone) {
      String sqlSelectStatement = 'DELETE FROM storage_node WHERE (path = ?);';
      try {
        _requestResult(sqlSelectStatement, [path]);
      } catch (e) {
        rethrow;
      }
    } else {
      Node deletedNode =
          await NodeImpl.fromPath(path, oldNode.owner, isTombstone: true);
      deletedNode.visibility = oldNode.visibility;
      await update(deletedNode);
    }
    if ('' != oldNode.parentPath) {
      Node parentNode = await get(oldNode.parentPath!);
      await parentNode.removeChild(oldNode.name);
      await update(parentNode);
    }
    return oldNode;
  }

  @override
  Future<NodeValue?> getValue(String path, String key) async {
    if (('' == path) || ('' == key)) {
      throw NullThrownError();
    }
    checkPath(path);
    Node ret = await get(path);
    if (ret.tombstone) {
      throw StorageException(
          'unable to get value... Node "$path" does no longer exist. Node is a tombstone.');
    }
    String sqlSelectStatement =
        ('SELECT path,key,value,type,locale,last_modified '
            'FROM node_value WHERE (path = ? and key = ?)');
    NodeValue? value;
    int lastModified = 0;
    try {
      ResultSet resultSet = _requestResult(sqlSelectStatement, [path, key]);
      if (resultSet.isEmpty) {
        return value;
      }
      lastModified = int.parse(resultSet.first['last_modified'] as String);
      value = NodeValueImpl(
          resultSet.first['key'] as String,
          resultSet.first['value'] as String,
          resultSet.first['type'] as String?,
          '',
          lastModified);
    } catch (e) {
      if (!(e is StorageException && e is SqliteException)) rethrow;
      throw StorageException(
          'Something went wrong while trying to retrieve the value ' +
              key +
              ' from ' +
              path,
          null,
          e);
    }
    String sqlStatementTranslations =
        ('SELECT path,key,identifier,locale,translation '
            'FROM translation WHERE (path = ? AND key = ?)');
    try {
      ResultSet rsTranslations =
          _requestResult(sqlStatementTranslations, [path, key]);
      for (Row rsTranslation in rsTranslations) {
        String identifier = rsTranslation['identifier'] as String;
        if ('VALUE' == identifier) {
          value.setValue(rsTranslation['translation'] as String,
              Locale.parse(rsTranslation['locale'] as String));
        } else {
          if ('DESCRIPTION' == identifier) {
            value.setDescription(rsTranslation['translation'] as String,
                Locale.parse(rsTranslation['locale'] as String));
          }
        }
      }
    } on SqliteException catch (e) {
      throw StorageException(
          'Could not retrieve description for node "' +
              path +
              '" and key "' +
              key +
              '"',
          null,
          e);
    }

    // make sure that lastModified is properly set
    value.lastModified = lastModified;
    return value;
  }

  Future<void> addValue(String path, NodeValue value) async {
    checkPath(path);
    await get(path);
    if (await getValue(path, value.key) != null) {
      throw StorageException('Value "${value.key}" already exists');
    }
    String sqlStatement =
        ('INSERT INTO node_value (path, key, value, type, locale, '
            'last_modified) VALUES (?,?,?,?,?,?)');
    try {
      _requestResult(sqlStatement, [
        path,
        value.key,
        value.value,
        value.type,
        NodeValueImpl.defaultLocale.toLanguageTag(),
        value.lastModified
      ]);
    } on StorageException catch (e) {
      throw StorageException('Could not create value "${value.key}"', null, e);
    }
    String sqlStatementTrl =
        ('INSERT INTO translation (path, key, identifier, locale, translation)'
            ' VALUES (?,?,?,?,?)');
    Map<Locale, String> valueMap = value.allValueTranslations;
    for (MapEntry<Locale, String> entry in valueMap.entries) {
      try {
        _requestResult(sqlStatementTrl,
            [path, value.key, 'VALUE', entry.key.toLanguageTag(), entry.value]);
      } on StorageException catch (e) {
        throw StorageException(
            'Could not create translation "' +
                entry.key.toLanguageTag() +
                '" for value "' +
                value.key +
                '"',
            null,
            e);
      }
    }
    Map<Locale, String> descriptionMap = value.allDescriptionTranslations;
    for (MapEntry<Locale, String> entry in descriptionMap.entries) {
      try {
        _requestResult(sqlStatementTrl, [
          path,
          value.key,
          'DESCRIPTION',
          entry.key.toLanguageTag(),
          entry.value
        ]);
      } on StorageException catch (e) {
        throw StorageException(
            'Could not create translation "' +
                entry.key.toLanguageTag() +
                '" for description in value "' +
                value.key +
                '"',
            null,
            e);
      }
    }
  }

  Future<void> updateValue(String path, NodeValue value) async {
    checkPath(path);
    if ('' == path) {
      throw NullThrownError();
    }
    if (await getValue(path, value.key) == null) {
      throw StorageException(('Key "' + value.key) + '" does not exist');
    }
    await deleteValue(path, value.key);
    await addValue(path, value);
  }

  Future<NodeValue> deleteValue(String path, String key) async {
    if ('' == path) {
      throw NullThrownError();
    }
    await get(path);
    NodeValue? value = await getValue(path, key);
    if (value == null) {
      throw StorageException(('Key "' + key) + '" does not exist');
    }
    String sqlDeleteStatementTranslations =
        'DELETE FROM translation WHERE (path = ? AND key = ?)';
    try {
      _requestResult(sqlDeleteStatementTranslations, [path, key]);
    } on StorageException catch (e) {
      throw StorageException(
          'Could not delete translations for key "' +
              key +
              '" in node "' +
              path +
              '"',
          null,
          e);
    }
    String sqlDeleteStatement =
        'DELETE FROM node_value WHERE (path = ? AND key = ?)';
    try {
      _requestResult(sqlDeleteStatement, [path, key]);
    } on StorageException catch (e) {
      throw StorageException(
          'Could not delete value for key "' + key + '" in node "' + path + '"',
          null,
          e);
    }
    return value;
  }

  @override
  Future<void> rename(String oldPath, String newPathOrName) async {
    checkPath(oldPath);
    checkPath(newPathOrName);
    // TODO(unassigned): check if new parentPath is not contained in old path
    Node oldNode = await get(oldPath);
    Node newNode;
    if (oldNode.tombstone) {
      // raise error if old node is a tombstone
      throw StorageException(
          'unable to rename node $oldPath... this node does not exist anymore');
    } else {
      String? path = NodeImpl.getNameFromPath(newPathOrName) == newPathOrName
          ? NodeImpl.getParentFromPath(oldPath)
          : NodeImpl.getParentFromPath(newPathOrName);
      newNode = NodeImpl(NodeImpl.getNameFromPath(newPathOrName)!,
          oldNode.owner, path, oldNode.visibility);
    }
    // copy all nodes to new node
    for (NodeValue nv in (await oldNode.getValues()).values) {
      await newNode.addValue(nv);
    }
    // create node at new location
    await add(newNode);

    // move all children
    for (Node n in (await oldNode.getChildren()).values) {
      await rename(
          n.path, newNode.path + GenericController.pathDelimiter + n.name);
    }

    // delete old node
    await delete(oldPath);
  }

  @override
  Future<List<Node>> search(SearchCriteria criteria) async {
    String sqlNodeSearch =
        'SELECT path, owner, name, visibility, children FROM storage_node '
        'WHERE (path LIKE ? and owner LIKE ? and name LIKE ? and visibility LIKE ?)';
    Map<String, Node> nodes = {};
    try {
      ResultSet resultSetNodes = _requestResult(sqlNodeSearch, [
        ('' == criteria.path) ? '%' : criteria.path + '%',
        (criteria.owner == null || criteria.owner == '') ? '%' : criteria.owner,
        (criteria.name == '') ? '%' : criteria.name,
        (criteria.visibility == '') ? '%' : criteria.visibility
      ]);
      for (Row node in resultSetNodes) {
        String path = node['path'] as String;
        Node tmp = await get(path);
        if (await criteria.evaluate(tmp)) {
          nodes[path] = tmp;
        }
      }
    } on SqliteException {
      rethrow;
    }
    return nodes.values.toList();
  }

  @override
  Future<void> close() async {
    try {
      conn.dispose();
    } on SqliteException catch (e) {
      throw Exception('OOPS... unexpected exception' + e.toString());
    }
  }

  @override
  Future<void> flush() async {
    // TODO(unassigned): implement flush
    /// try {
    //conn.commit();
    /// } catch (SQLException e) {
    /// throw new RuntimeException("Could not flush database", e);
    /// }
  }

  @override
  Future<void> zap() async {
    String sqlStatement1 = 'DELETE FROM storage_node';
    String sqlStatement2 = 'DELETE FROM node_value';
    String sqlStatement3 = 'DELETE FROM translation';
    try {
      conn.execute(sqlStatement3);
      conn.execute(sqlStatement2);
      conn.execute(sqlStatement1);
    } on SqliteException catch (e) {
      throw StorageException(
          'Something went wrong while trying to truncate the database',
          null,
          e);
    }
  }

  /// Executes the given sqlStatement as a prepared statement using the optional args in given order.
  ///
  /// @param sqlStatement the statement to execute
  /// @param args the arguments to replace
  /// @return the result set containing all the results
  ResultSet _requestResult(String sqlStatement, List<Object?> args) {
    try {
      return conn.select(sqlStatement, args);
    } on SqliteException catch (e, st) {
      throw StorageException(
          'Could not execute query: "$sqlStatement" with args: ${args.toString()} [${e.toString()}]',
          null,
          e,
          st);
    }
  }
}
