library geiger_localstorage;

import 'dart:io';

import 'package:flutter/services.dart';
import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:path_provider/path_provider.dart';

/// <p>Generic interface to define the common methods between databases.</p>
///
/// @author Sacha
abstract class StorageMapper {
  static String? _databasesPath;

  static bool preferCurrentDirectory = true;

  /// Sets the controller to be used when creating skeletons.</p>
  ///
  /// @param controller the controller to be used to resolve skeletons
  void setController(StorageController controller);

  /// <p>Get a node by node name from the storage backend.</p>
  ///
  /// @param path the fully qualified node name
  /// @return the requested node
  /// @throws StorageException if node is not found or an error in the storage API happens
  Future<Node> get(String path);

  /// <p>Add a non existing node to the storage backend.</p>
  ///
  /// @param node the node to be added to the backend
  /// @throws StorageException if the node already exists, the parental node does not exist,
  /// or the backend storage encountered a problem
  Future<void> add(Node node);

  /// <p>Update an already existing node with the current values.</p>
  ///
  /// <p>
  /// All current values of the passed on node are written into the new node. Any non existing
  /// values are removed.
  /// </p>
  /// @param node the node with the values to be written
  /// @throws StorageException if the node does not exist or an error happened in the storage
  /// backend.
  Future<void> update(Node node);

  /// <p>Remove the named node.</p>
  ///
  /// @param path the fully qualified path of the node to be removed
  /// @return A representation of the removed node
  /// @throws StorageException if the node does not exist, the node contains at least one child,
  /// or an error happened on the backend
  Future<Node> delete(String path);

  /// <p>Fetch a value from a node.</p>
  ///
  /// @param path the fully qualified path to a node.
  /// @param key  the key to be fetched
  /// @return a representation of the node value
  /// @throws StorageException if the storage backend encounters a problem
  Future<NodeValue?> getValue(String path, String key);

  /// <p>Renames or moves an existing node.</p>
  ///
  /// @param oldPath       the old path of the node as fully qualified name
  /// @param newPathOrName the new path of the node as fully qualified path or the new name of the
  /// node
  /// @throws StorageException if the new node already exists, the old node does not exist, the new
  /// parent node does not exist or the storage backend encountered
  /// problems
  Future<void> rename(String oldPath, String newPathOrName);

  /// <p>Search all nodes matching the provided search criteria.</p>
  ///
  /// @param criteria the matching criteria (@see ch.fhnw.geiger.localstorage.SearchCriteria for
  /// specification of search)
  /// @return the nodes matching the criteria
  /// @throws StorageException if the storage backend encounters a problem
  Future<List<Node>> search(SearchCriteria criteria);

  /// <p>closes the database backend and flushes all data.</p>
  Future<void> close();

  /// <p>Flushes all data possible to the disk.</p>
  Future<void> flush();

  /// <p>Zaps the current database and discards all values.</p>
  ///
  /// @throws StorageException In case of problems regarding the storage
  Future<void> zap();

  /// Extend database path name with writeable base dir if no base dir is specified
  static String expandDbFileName(String name) {
    if (!name.startsWith('.${Platform.pathSeparator}') &&
        !name.startsWith(Platform.pathSeparator)) {
      // Get a location using getDatabasesPath
      if (_databasesPath == null) {
        throw StorageException(
            'Database path not yet known... initialize first with "await StorageMapper.initDatabaseExpander()"');
      }
      String path = '$_databasesPath${Platform.pathSeparator}$name';
      return path;
    } else {
      return name;
    }
  }

  static bool isWriteable([String path = "."]) {
    bool res;
    final f = File('$path${Platform.pathSeparator}test.tst');
    final didExist = f.existsSync();
    try {
      // try appending nothing
      f.writeAsStringSync('', mode: FileMode.append, flush: true);
      res = true;
      if (didExist) {
        f.deleteSync();
      }
    } on FileSystemException {
      res = false;
    }
    return res;
  }

  static Future<void> initDatabaseExpander() async {
    Directory directory;
    // use path provider only if the current directory is not writeable
    if (!(isWriteable('.') && preferCurrentDirectory)) {
      if (Platform.isMacOS) {
        // MacOS path_provider implementation was too stupid (unimplemented or null operator problem)
        directory = Directory('~/Library/Application Support');
      } else if (Platform.isWindows || Platform.isLinux || Platform.isAndroid) {
        directory = await getApplicationSupportDirectory();
      } else if (Platform.isAndroid) {
        directory = await getApplicationDocumentsDirectory();
      } else {
        try {
          directory = await getLibraryDirectory();
        } on MissingPluginException {
          // fall back to the current directory
          directory = Directory.current;
        }
      }
    } else {
      directory = Directory.current;
    }
    _databasesPath = directory.path;
  }
}
