// ignore_for_file: avoid_print

library geiger_localstorage;

import 'dart:convert';

import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:test/test.dart';

class SerializerHelperTest {}

void main() {
  group('Testing byte streams and sinks', () {
    test('sink test String', () async {
      for (String str in ['hello', 'öä\$üç&+"@']) {
        ByteSink s = ByteSink();
        s.sink.add(utf8.encode(str));
        s.close();
        expect(str, utf8.decode(await s.bytes), reason: 'strings not equal');
      }
    });
    test('sink test list<int>', () async {
      for (List<int> str in [
        [0],
        [-1, 1]
      ]) {
        ByteSink s = ByteSink();
        s.sink.add(str);
        s.close();
        expect(str, await s.bytes, reason: 'lists not equal');
      }
    });
  });
  group('primitive de/serialization tests', () {
    test('testing byte order serialization', () async {
      ByteSink sink = ByteSink();
      SerializerHelper.writeInt(sink, 257);
      sink.close();
      var l = await sink.bytes;
      expect(l[0], 89, reason: 'uid byte 1 mismatch');
      expect(l[1], 159, reason: 'uid byte 2 mismatch');
      expect(l[2], 201, reason: 'uid byte 3 mismatch');
      expect(l[3], 111, reason: 'uid byte 4 mismatch');
      expect(l[4], 28, reason: 'uid byte 5 mismatch');
      expect(l[5], 0, reason: 'uid byte 6 mismatch');
      expect(l[6], 0, reason: 'uid byte 7 mismatch');
      expect(l[7], 0, reason: 'uid byte 8 mismatch');
      expect(l[8], 1, reason: 'data byte 1 mismatch');
      expect(l[9], 1, reason: 'data byte 2 mismatch');
      expect(l[10], 0, reason: 'data byte 3 mismatch');
      expect(l[11], 0, reason: 'data byte 4 mismatch');
    });
    test('testing serialization/deserialization of int', () async {
      for (int s in [0, 1, -1, (1 << 31) - 1]) {
        ByteSink sink = ByteSink();
        SerializerHelper.writeInt(sink, s);
        sink.close();
        var l = await sink.bytes;
        expect(12, l.length, reason: 'Length mismatch of stream');
        //print('## number=${s}');
        //print('## 0=${l[8]}');
        //print('## 1=${l[9]}');
        //print('## 2=${l[10]}');
        //print('## 3=${l[11]}');
        //print('');
        var stream = ByteStream(null, l);
        var s2 = await SerializerHelper.readInt(stream);
        expect(s2, s,
            reason: 'deserialization of int failed for $s (resulted in $s2)');
        expect((await stream.bytes).length, 0,
            reason: 'deserialization did not empty stream');
      }
    });
    test('testing serialization/deserialization of long', () async {
      for (int s in [0, 1, -1, (1 << 63) - 1]) {
        ByteSink sink = ByteSink();
        SerializerHelper.writeLong(sink, s);
        sink.close();
        var stream = ByteStream(null, await sink.bytes);
        var s2 = await SerializerHelper.readLong(stream);
        expect(s2, s, reason: 'deserialization of int failed for $s');
        expect((await stream.bytes).length, 0,
            reason: 'deserialization did not empty stream');
      }
    });
    test('testing serialization/deserialization of string', () async {
      for (String s in ['', 'string', 'testMessage öäü^']) {
        ByteSink sink = ByteSink();
        SerializerHelper.writeString(sink, s);
        sink.close();
        ByteStream stream = ByteStream(null, await sink.bytes);
        var s2 = await SerializerHelper.readString(stream);
        expect(s2, s, reason: 'deserialization failed for string "$s"');
        expect((await stream.bytes).length, 0,
            reason: 'deserialization did not empty stream');
      }
    });
  });
  group('exception de/serialization tests', () {
    test('testing serialization/deserialization of SerializedException',
        () async {
      var stackTraces = [null, StackTrace.empty, StackTrace.current];
      var exceptions = [
        SerializedException(''),
        SerializedException('name'),
        SerializedException('', message: 'message'),
        ...stackTraces
            .map((trace) => SerializedException('', stackTrace: trace)),
        SerializedException('', causeStackTrace: StackTrace.current),
      ];
      var causes = [Exception(), Exception('name'), ...exceptions];
      exceptions.addAll(causes.expand((cause) => stackTraces.map((trace) =>
          SerializedException('', cause: cause, stackTrace: trace))));

      checkException(SerializedException e1, SerializedException e2) {
        expect(e1.name, e2.name,
            reason: 'names "${e1.name}" and "${e2.name}" are not equal');
        expect(e1.message, e2.message,
            reason:
                'messages "${e1.message}" and "${e2.message}" are not equal');
        expect(e1.stackTrace.toString(), e2.stackTrace.toString(),
            reason: 'stack traces are not equal');
        if (e1.cause == null) {
          expect(e2.cause, null, reason: 'cause is not null');
        } else if (e2.cause == null) {
          fail('cause is unexpectedly null');
        } else {
          checkException(
              e1.cause as SerializedException, e2.cause as SerializedException);
        }
      }

      for (SerializedException e1 in exceptions) {
        ByteSink sink = ByteSink();
        e1.toByteArrayStream(sink);
        sink.close();
        var e2 = await SerializedException.fromByteArrayStream(
            ByteStream(null, await sink.bytes));
        checkException(e1, e2);
      }
    });
    test('testing serialization/deserialization of StorageException', () async {
      var stackTraces = [null, StackTrace.empty, StackTrace.current];
      var exceptions = [
        StorageException(''),
        StorageException('name'),
        ...stackTraces.map((trace) => StorageException('', trace)),
        StorageException('', StackTrace.current),
      ];
      var causes = [Exception(), Exception('name'), ...exceptions];
      exceptions.addAll(causes.expand(
          (cause) => stackTraces.map((trace) => StorageException('', trace))));

      void checkException(SerializedException e1, SerializedException e2) {
        expect(e1.name, e2.name,
            reason: 'names "${e1.name}" and "${e2.name}" are not equal');
        expect(e1.message, e2.message,
            reason:
                'messages "${e1.message}" and "${e2.message}" are not equal');
        expect(e1.stackTrace.toString(), e2.stackTrace.toString(),
            reason: 'stack traces are not equal');
        if (e1.cause == null) {
          expect(e2.cause, isNull, reason: 'cause is not null');
        } else if (e2.cause == null) {
          fail('cause is unexpectedly null');
        } else {
          checkException(
              e1.cause as StorageException, e2.cause as StorageException);
        }
      }

      for (SerializedException e1 in exceptions) {
        ByteSink sink = ByteSink();
        e1.toByteArrayStream(sink);
        sink.close();
        var e2 = await SerializedException.fromByteArrayStream(
            ByteStream(null, await sink.bytes));
        print("Testing exception \n$e1\nagainst\n$e2");
        checkException(e1, e2);
      }
    });
  });
  group('multi read/write tests', () {
    test('testing int', () async {
      ByteSink sink = ByteSink();
      SerializerHelper.writeInt(sink, 1);
      SerializerHelper.writeInt(sink, 2);
      sink.close();
      ByteStream stream = ByteStream(null, await sink.bytes);
      expect(await SerializerHelper.readInt(stream), 1);
      expect(await SerializerHelper.readInt(stream), 2);
    });
    test('testing long', () async {
      ByteSink sink = ByteSink();
      SerializerHelper.writeLong(sink, 1);
      SerializerHelper.writeLong(sink, -1);
      sink.close();
      ByteStream stream = ByteStream(null, await sink.bytes);
      expect(await SerializerHelper.readLong(stream), 1);
      expect(await SerializerHelper.readLong(stream), -1);
    });
    test('testing String', () async {
      ByteSink sink = ByteSink();
      SerializerHelper.writeString(sink, null);
      SerializerHelper.writeString(sink, '');
      SerializerHelper.writeString(sink, 'abcdefg');
      sink.close();
      ByteStream stream = ByteStream(null, await sink.bytes);
      expect(await SerializerHelper.readString(stream), null);
      expect(await SerializerHelper.readString(stream), '');
      expect(await SerializerHelper.readString(stream), 'abcdefg');
    });
  });
}
