// ignore_for_file: avoid_print, prefer_const_constructors

import 'package:geiger_localstorage/geiger_localstorage.dart';
import 'package:intl/locale.dart';
import 'package:test/test.dart';

class StorageMapperTest {}

void _testPrep(List<StorageMapper> mapperList) {
  for (final mapper in mapperList) {
    mapper.zap();
  }
}

void _testTearDown(List<StorageMapper> mapperList) {
  for (final mapper in mapperList) {
    mapper.flush();
  }
}

void main() {
  late List<StorageMapper> mapperList;

  setUp(() async {
    await StorageMapper.initDatabaseExpander();
    mapperList = [DummyMapper('testowner'), SqliteMapper('./testdb.sqlite')];
    _testPrep(mapperList);
  });

  test('Adding node', () async {
    for (final mapper in mapperList) {
      print('## Testing mapper $mapper in UNKNOWN');
      Node node = NodeImpl('testNode1', 'testowner', ':');
      node.owner = 'myOwner';
      // make sure root node exists
      try {
        await mapper.add(NodeImpl(':', 'testowner'));
      } on StorageException {
        /*save to ignore*/
      }

      print("### adding new node");
      await mapper.add(node);

      // make sure that lastupdated and owner is not updated upon add
      print("### retrieving previously written node");
      Node node2 = await mapper.get(':testNode1');
      print("### testing node content");
      expect(node2.equals(node), true,
          reason: 'Last updated was unexpectedly updated');
    }
  });

  test('add root node null', () async {
    for (final mapper in mapperList) {
      print('## Testing mapper ' + mapper.toString() + ' in UNKNOWN');
      Node node = NodeImpl(':testNode1', 'testowner');
      node.owner = 'myOwner';

      // make sure root node exists
      try {
        await mapper.add(NodeImpl(':', ''));
      } on StorageException {
        /*save to ignore*/
      }
      await mapper.add(node);

      // make sure that lastupdated and owner is not updated upon add
      Node node2 = await mapper.get(':testNode1');
      expect(node2.equals(node), true,
          reason: 'Last updated was unexpectedly updated');
    }
  });

  test('add root node illegal name', () async {
    for (final StorageMapper mapper in mapperList) {
      for (String name in ['na:me', 'name:']) {
        print('## Testing mapper ' + mapper.toString() + ' in UNKNOWN');
        expect(() => NodeImpl(name, 'testowner'),
            throwsA(TypeMatcher<StorageException>()));
      }
    }
  });

  test('add root node illegal parent', () async {
    for (final StorageMapper mapper in mapperList) {
      // make sure root node exists
      try {
        await mapper.add(NodeImpl(':', ''));
      } on StorageException {
        /*save to ignore*/
      }

      await mapper.add(NodeImpl(':na', 'owner'));
      for (var parent in [':name:', 'na:me', 'name:', '::', ':na::']) {
        print('## Testing mapper ' + mapper.toString() + ' in INVALID');
        expect(() async => NodeImpl('name', 'testowner', parent),
            throwsA(TypeMatcher<StorageException>()));
      }
    }
  });
  test('add duplicate node', () async {
    for (final StorageMapper mapper in mapperList) {
      // make sure root node exists
      try {
        await mapper.add(NodeImpl(':', ''));
      } on StorageException {
        /*save to ignore*/
      }

      print('## Testing mapper ' + mapper.toString() + ' in INVALID');
      Node node = NodeImpl(':testNode1', 'o');
      await mapper.add(node);
      expect(() async => await mapper.add(node),
          throwsA(TypeMatcher<StorageException>()));
    }
  });
  test('add parent-less node', () async {
    for (final StorageMapper mapper in mapperList) {
      // make sure root node exists
      try {
        await mapper.add(NodeImpl(':', ''));
      } on StorageException {
        /*save to ignore*/
      }

      print('## Testing mapper ' + mapper.toString() + ' in INVALID');
      Node node = NodeImpl('testNode1', 'testOwner', ':notAValidParent');
      expect(() async => await mapper.add(node),
          throwsA(TypeMatcher<StorageException>()));
    }
  });
  test('Get non exiting nodeValue', () async {
    for (final StorageMapper mapper in mapperList) {
      mapper.zap();
      // make sure root node exists
      try {
        await mapper.add(NodeImpl(':', ''));
      } on StorageException {
        /*save to ignore*/
      }

      print('## Testing mapper ${mapper.toString()}');
      Node node = NodeImpl(':testNode1', 'testOwner');
      await mapper.add(node);
      NodeValue? nv = await mapper.getValue(':testNode1', 'nonexistentKey');
      expect(nv, null,
          reason:
              'Mapper $mapper did not return null value for nonexistent key');
    }
  });
  test('node getter tests', () async {
    for (final mapper in mapperList) {
      // make sure root node exists
      try {
        await mapper.add(NodeImpl(':', ''));
      } on StorageException {
        /*save to ignore*/
      }

      print('## Testing mapper ' + mapper.toString() + ' in INVALID');
      Node node = NodeImpl('testNode1', 'testOwner', ':');
      Node childNode = NodeImpl('testNode1a', 'testOwner', ':testNode1');
      NodeValue nv = NodeValueImpl('key', 'value', 'type', 'description', 1);
      await node.addChild(childNode);
      await node.addValue(nv);

      // write data
      Node node2 = NodeImpl('testNode2', 'testOwner', ':');
      await mapper.add(node);
      await mapper.add(node2);
      await mapper.add(childNode);

      // get data
      Node storedNode = await mapper.get(':testNode1');
      Node storedChildNode = await mapper.get(':testNode1:testNode1a');
      Node storedNode2 = await mapper.get(':testNode2');
      expect(storedNode.skeleton, false,
          reason: 'Node :testNode1 may not be a skeleton if fetched');
      expect(storedChildNode.skeleton, false,
          reason: 'Node :testNode1 may not be a skeleton if fetched');
      expect(storedNode2.skeleton, false,
          reason: 'Node :testNode1 may not be a skeleton if fetched');

      // compare data
      expect(node.toString(showTimestamp: false),
          storedNode.toString(showTimestamp: false),
          reason:
              'comparing parent node failed (expected: $node; got: $storedNode');
      expect(childNode.equals(storedChildNode), isTrue,
          reason: 'comparing child node');
      expect(node2.equals(storedNode2), isTrue,
          reason: 'comparing parent node2 ');
      expect((await storedNode.getChildren()).length, 1,
          reason: 'checking for child node count');
    }
  });
  test('persistence of locales', () async {
    for (final StorageMapper mapper in mapperList) {
      // make sure root node exists
      try {
        await mapper.add(NodeImpl(':', ''));
      } on StorageException {
        /*save to ignore*/
      }

      print('## Testing mapper ' + mapper.toString() + ' in INVALID');
      var nv = NodeValueImpl('key', 'value', 'type', 'description', 1);
      nv.setDescription('Deutsche Beschreibung', Locale.parse('de-de'));
      nv.setDescription('Schwiizertüütschi Beschriibig', Locale.parse('de-ch'));
      nv.setValue('Wert', Locale.parse('de-de'));
      nv.setValue('Au ä Wert', Locale.parse('de-ch'));
      Node node = NodeImpl('testNode1', 'testOwner', ':');
      Node childNode = NodeImpl('testNode1a', 'testOwner', ':testNode1');
      await node.addChild(childNode);
      await node.addValue(nv);
      print('## Stored ' + node.toString());

      // write data
      Node node2 = NodeImpl('testNode2', 'testOwner', ':');
      await mapper.add(node);
      await mapper.add(node2);
      await mapper.add(childNode);

      // get data
      Node storedNode = await mapper.get(':testNode1');
      Node storedChildNode = await mapper.get(':testNode1:testNode1a');
      Node storedNode2 = await mapper.get(':testNode2');
      expect(storedNode.skeleton, false,
          reason: 'Node :testNode1 may not be a skeleton if fetched');
      expect(storedChildNode.skeleton, false,
          reason:
              'Node :testNode1:testNode1a may not be a skeleton if fetched');
      expect(storedNode2.skeleton, false,
          reason: 'Node :testNode2 may not be a skeleton if fetched');
      print('## got    ' + storedNode.toString());

      // compare data
      expect(node.toString(showTimestamp: false),
          storedNode.toString(showTimestamp: false),
          reason:
              'comparing parent node (expected: ${node.toString()}(${node.toString().hashCode}); got: ${storedNode.toString()}(${storedNode.toString().hashCode})');
      expect(childNode.equals(storedChildNode), true,
          reason: 'comparing child node');
      expect(node2.equals(storedNode2), true, reason: 'comparing parent node2');
      expect((await storedNode.getChildren()).length, 1,
          reason: 'checking for child node count');
    }
  });
  test('persistence of node timestamps', () async {
    for (final StorageMapper mapper in mapperList) {
      // make sure root node exists
      try {
        await mapper.add(NodeImpl(':', ''));
      } on StorageException {
        /*save to ignore*/
      }

      String nn = ':timestampMapperPersistenceTest';
      try {
        await mapper.delete(nn);
      } on StorageException {
        // ignore this exception
      }
      Node n = NodeImpl(nn, 'test');
      await mapper.add(n);

      Node n2 = await mapper.get(nn);
      expect(n.lastModified, n2.lastModified,
          reason: 'inequal timestamp with mapper $mapper');
    }
  });
  test('check for child node enumeration', () async {
    for (final StorageMapper mapper in mapperList) {
      await mapper.zap();

      await mapper.add(NodeImpl(':', 'testowner'));
      for (int i = 0; i < 6; i++) {
        Node n = NodeImpl(':node$i', 'testowner');
        await mapper.add(n);
      }
      Node n = await mapper.get(':');
      Map<String, Node> nl = await n.getChildren();
      expect(nl.length > 5, isTrue,
          reason: 'too few child nodes  with mapper $mapper (${nl.length})');
      String s = await n.getChildNodesCsv();
      expect(s.split(',').length, nl.length,
          reason: 'inequal number of child nodes with mapper $mapper');
    }
  });

  tearDown(() async {
    _testTearDown(mapperList);

    // tear down all test data
    for (final StorageMapper mapper in mapperList) {
      await mapper.close();
    }
    // File('testdb.sqlite').deleteSync();
  });
}
